import time
import waiting

from fild_compare import compare
from fild_cfg import Cfg
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common import action_chains
from selenium.common.exceptions import (
    ElementClickInterceptedException, StaleElementReferenceException,
    WebDriverException,
)
from selenium.webdriver import Chrome, Remote
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from waiting import wait, TimeoutExpired

from fild_ui import folder
from fild_ui.screenshots import PageScreenshot
from fild_ui.singleton import Singleton


def retry_on_js_reload(function, refresh, timeout_seconds=2):
    result = None

    def action():
        try:
            nonlocal result
            result = function()
            return True
        except StaleElementReferenceException:
            refresh()
            return None
        except ElementClickInterceptedException:
            refresh()
            return None

    waiting.wait(
        action,
        timeout_seconds=timeout_seconds,
        waiting_for='the elements to load'
    )
    return result


class Browser(metaclass=Singleton):
    _driver = None
    _session_id = 0

    @property
    def driver(self):
        if self._driver is None:
            options = Options()

            options.add_argument('--allow-insecure-localhost')
            options.add_argument('--ignore-certificate-errors')
            options.add_argument('--disable-web-security')
            options.add_argument('--disable-gpu')
            options.add_argument('--hide-scrollbars')
            options.add_argument('--disable-site-isolation-trials')
            options.add_argument('--remote-debugging-pipe')

            if Cfg.Browser.headless:
                options.add_argument('--headless=new')

            if Cfg.Browser.get('devtools'):
                options.add_argument('--auto-open-devtools-for-tabs')

            if Cfg.Browser.get('no_sandbox'):
                options.add_argument('--no-sandbox')
                options.add_argument('--disable-setuid-sandbox')
                options.add_argument('--disable-dev-shm-usage')

            options.add_experimental_option('prefs', {
                'download.default_directory':
                    str(folder.generate_path('downloads').absolute()),
                'pageLoadStrategy': 'none',
            })

            if Cfg.Browser.get('remote_url'):
                self._driver = Remote(
                    options=options, command_executor=Cfg.Browser.remote_url
                )
            else:
                self._driver = Chrome(options=options)

            self._driver.set_page_load_timeout(15)
            self._driver.set_window_size(1680, 1050)

        return self._driver

    def close(self):
        if self._driver is None:
            return

        self._driver.quit()
        self._driver = None

    @classmethod
    def recreate_session(cls):
        cls._session_id += 1

    @property
    def is_open(self):
        return self._driver is not None

    @property
    def session(self):
        return self._session_id

    @property
    def console_log(self):
        logs = self.driver.get_log('browser')
        severe_logs = [log for log in logs if log['level'] == 'SEVERE']

        return severe_logs


class Page:
    base_url = None
    url = None

    @classmethod
    def open(cls, *args, url_postfix=None):
        cls.open_page(*args, url_postfix=url_postfix)

    @classmethod
    def open_page(cls, *args, url_postfix=None):
        if cls.base_url is None and cls.url is None:
            raise NotImplementedError('Page url is not specified')

        url = f'{cls.base_url}/{(cls.url or "").format(*args)}{url_postfix or ""}'
        Browser().recreate_session()
        Browser().driver.get(url)

    @classmethod
    def scroll_to_bottom(cls):
        Browser().driver.execute_script(
            "window.scrollTo(0,document.body.scrollHeight);"
        )
        Browser().driver.execute_script(
             "document.querySelector('#root').scrollTo(0,document.body.scrollHeight);"
        )
        # TODO make it work w/o sleep
        time.sleep(0.5)

    @classmethod
    def get_current_url(cls):
        return Browser().driver.current_url

    @classmethod
    def wait_for_redirect(cls):
        waiting.wait(
            lambda: '?_=' in cls.get_current_url(),
            timeout_seconds=1,
            waiting_for='url to update'
        )

    @classmethod
    def wait_for_current_url(cls, url):
        waiting.wait(
            lambda: cls.get_current_url() == url,
            timeout_seconds=1,
            waiting_for='url to update'
        )

    @classmethod
    def is_open(cls, *args, url_postfix=None, skip_params=True):
        actual = cls.get_current_url()

        if skip_params:
            actual = actual.split('?')[0]

        page_url = (
            f'{cls.base_url}/{(cls.url or "").format(*args)}'
            f'{url_postfix or ""}'
        )

        return page_url == actual

    @classmethod
    def verify_current_url(cls, *args, url_postfix=None, skip_params=True):
        actual = cls.get_current_url()

        if skip_params:
            actual = actual.split('?')[0]

        expected = (
            f'{cls.base_url}/{(cls.url or "").format(*args)}'
            f'{url_postfix or ""}'
        )

        assert actual == expected, (
            f'Unexpected url.\nExpected: {expected}\nGot: {actual}'
        )

    @classmethod
    def go_back(cls):
        return Browser().driver.back()

    @staticmethod
    def press_key(key):
        action_chains.ActionChains(Browser().driver).send_keys(key).perform()

    @staticmethod
    def wait_for_window_closed(timeout=5):
        wait(
            lambda: len(Browser().driver.window_handles) == 1,
            timeout_seconds=timeout,
            waiting_for='extra windows to close'
        )

    @staticmethod
    def switch_to_default_window():
        Browser().driver.switch_to.window(
            Browser().driver.window_handles[0]
        )


class Element:
    _fixed_target = None
    _browser_session = None
    _parent = None
    _target = None

    _height = None
    _width = None

    def __init__(self, by_value=None, by=None, # pylint: disable=invalid-name
                 css=None, xpath=None, name=None):
        if css:
            by = By.CSS_SELECTOR
            by_value = css
        elif xpath:
            by = By.XPATH
            by_value = xpath
        elif name:
            by = By.NAME
            by_value = name

        self._fixed_target = None
        self.by = by or By.ID  # pylint: disable=invalid-name
        self.by_value = by_value

    def is_list_item_container(self):
        return self._fixed_target is not None

    @classmethod
    def set_located(cls, target):
        element = cls(by_value=None)
        element._fixed_target = target

        return element

    @property
    def parent(self):
        if self._parent is None:
            return Browser().driver

        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent

    def invalidate(self):
        self._target = None

        if self._parent:
            self._parent.invalidate()

    def check_browser_session(self):
        actual_session = Browser().session

        if self._browser_session != actual_session:
            self.invalidate()
            self._browser_session = actual_session

    def find_element(self, by, value):
        return self.located.find_element(by=by, value=value)

    def find_elements(self, by, value):
        return self.located.find_elements(by=by, value=value)

    @property
    def located(self):
        if self._fixed_target is not None:
            return self._fixed_target

        self.check_browser_session()

        if self._target is None:
            self._target = self.parent.find_element(
                by=self.by, value=self.by_value
            )

        return self._target

    def get_attribute(self, name):
        return self.located.get_attribute(name=name)

    def __get__(self, instance, owner):
        if instance and issubclass(owner, Element):
            if instance.is_list_item_container:
                Browser().recreate_session()

            self.parent = instance

        return self

    def __getattr__(self, item):
        if hasattr(self.__class__, item):
            return object.__getattribute__(self, item)

        return self.located.__getattribute__(item)

    @property
    def text(self):
        return self.located.text

    def wait_for_load(self, timeout_seconds=2):
        wait(lambda: self.is_present(timeout=1),
             timeout_seconds=timeout_seconds,
             waiting_for='item to be displayed')

    def wait_for_not_present(self, timeout_seconds=1):
        wait(lambda: self.is_present(timeout=0) is False,
             timeout_seconds=timeout_seconds,
             sleep_seconds=0,
             waiting_for='item to be not present')

    def is_not_present(self, timeout_seconds=1):
        try:
            wait(lambda: self.is_present(timeout=0) is False,
                 timeout_seconds=timeout_seconds,
                 waiting_for='item to be not displayed')
            return True
        except TimeoutExpired:
            return False

    def is_present(self, timeout=0):
        def wait_for_is_displayed():
            Browser().recreate_session()
            return self.located.is_displayed()

        Browser().driver.implicitly_wait(timeout)

        try:
            wait(wait_for_is_displayed, timeout_seconds=timeout)
            return True
        except WebDriverException:
            return False
        except TimeoutExpired:
            return False
        finally:
            Browser().driver.implicitly_wait(1)

    def scroll_to(self, timeout=1):
        self.wait_for_load()
        viewport_height = Browser().driver.get_window_size()['height']
        height = self.located.size['height']

        if self.located.location['y'] + height <= viewport_height:
            return

        wait(
            lambda: self.located.location_once_scrolled_into_view['y'] + height <= viewport_height,
            timeout_seconds=timeout,
            waiting_for='element to be scrolled into view'
        )

    def delete(self):
        self.wait_for_load()

        Browser().driver.execute_script(
            'arguments[0].parentNode.removeChild(arguments[0])',
            self.located
        )

    def save_screenshot(self, file_name):
        self.located.screenshot(filename=file_name)

    def get_screenshot_as_png(self):
        return self.located.screenshot_as_png

    def wait_for_size(self, height=None, width=None):
        def is_of_size():
            Browser().recreate_session()

            if height is not None:
                if self.located.size['height'] != height:
                    return False

            if width is not None:
                if self.located.size['width'] != width:
                    return False

            return True

        wait(
            is_of_size,
            sleep_seconds=0,
            timeout_seconds=2,
            waiting_for=(
                f'element to load to size {height}x{width}. '
                f'Latest was: {self.located.size}'
            )
        )

    def verify_by_screenshot(self):
        PageScreenshot.compare(self)

    def verify_displayed(self, height=None, width=None):
        self.scroll_to()
        self.wait_for_size(
            height=height or self._height,
            width=width or self._width
        )

        if Cfg.Screenshot.compare:
            self.verify_by_screenshot()
        else:
            assert self.located.is_displayed()

    def hover(self):
        self.wait_for_load()
        ActionChains(Browser().driver).move_to_element(self.located).perform()

        return True  # for syncronization in waits


class Elements(Element):
    def __init__(self, by_value=None, by=None, css=None, xpath=None,
                 element_class=None):
        super().__init__(by_value=by_value, by=by, css=css, xpath=xpath)
        self.element_class = element_class

    @property
    def located(self):
        self.check_browser_session()

        if self._target is None:
            self._target = self.parent.find_elements(
                by=self.by, value=self.by_value
            )

        if self.element_class:
            self._target = [self.element_class.set_located(element)
                            for element in self._target]

        return self._target

    def reload(self, parent):
        if self._parent:
            self._parent.invalidate()
            self._target = self.parent.find_elements(by=self.by, value=self.by_value)
        else:
            self._target = parent.find_elements(by=self.by, value=self.by_value)

        if self.element_class:
            self._target = [self.element_class.set_located(element)
                            for element in self._target]

    def is_present(self, timeout=0):
        Browser().driver.implicitly_wait(0)
        try:
            wait(self.located[0].is_displayed, timeout_seconds=timeout)
            return True
        except WebDriverException:
            return False
        except TimeoutExpired:
            return False
        except IndexError:
            return False
        finally:
            Browser().driver.implicitly_wait(1)

    def __len__(self):
        return len(self.located)

    def __getitem__(self, item):
        Browser().recreate_session()
        return self.located[item]  # Return Field to provide attribute access

    def __iter__(self):
        return iter(self.located)

    def get_by_attribute(self, **kwargs):
        name = next(iter(kwargs))
        value = kwargs[name]

        for element in self.located:
            if element.get_attribute(name) == value:
                return element

        return None

    def get_by_text(self, text, strip=False):
        for element in self.located:
            current = element.text

            if strip:
                current = current.strip()

            if current == text:
                return element

        return None

    def wait_for_items_load(self, items_count, timeout_seconds=2):
        def check_count():
            if not items_count:
                Browser().driver.implicitly_wait(0)

            Browser().recreate_session()
            return len(self.located) == items_count

        try:
            wait(check_count,
                 timeout_seconds=timeout_seconds,
                 waiting_for=f'{items_count} items. Got: {len(self.located)}')
        finally:
            Browser().driver.implicitly_wait(1)

    def wait_for_some_items(self, timeout_seconds=2):
        def check_count():
            Browser().recreate_session()
            return len(self.located) != 0

        wait(check_count,
             timeout_seconds=timeout_seconds,
             waiting_for='some items to load')

    def get_labels(self):
        return [item.text for item in self.located]

    def verify_labels(self, expected):
        compare(
            expected=expected,
            actual=self.get_labels(),
            target_name='elements labels'
        )



import os

from pathlib import Path
from datetime import datetime

from waiting import wait


def generate_path(folder_name):
    path = Path(folder_name)

    try:
        if not os.path.exists(path):
            os.makedirs(path)
    except FileExistsError:
        # it's race condition
        pass

    return path


def generate_file_name(postfix='', extension='png'):
    timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')

    return f'{timestamp}_{postfix}.{extension}'


def generate_file_path(folder_name, filename):
    path = generate_path(folder_name)

    return str(path / filename)


def get_downloaded_file_with_wait(extension=None):
    downloads_folder = generate_path('downloads').absolute()

    def wait_for_file():
        files = os.listdir(downloads_folder)

        for file_name in files:
            if extension is None or file_name.endswith(extension):
                return True

        return False

    wait(
        wait_for_file,
        timeout_seconds=2,
        waiting_for='file to download'
    )

    for filename in os.listdir(downloads_folder):
        return str(os.path.join(downloads_folder, filename))



import base64

import json

from fild_compare import compare
from selenium.common.exceptions import WebDriverException

from fild_ui.browser import Browser


class Command:
    GET_ITEM = 'return window.localStorage.getItem(arguments[0]);'
    ADD_ITEM = 'window.localStorage.setItem(arguments[0], arguments[1]);'
    DELETE = 'window.localStorage.removeItem(arguments[0]);'
    CLEAR = 'window.localStorage.clear();'


class LocalStorage:
    @classmethod
    def run_command(cls, command, *args):
        return Browser().driver.execute_script(command, *args)

    @classmethod
    def get_item(cls, key):
        return cls.run_command(Command.GET_ITEM, key)

    @classmethod
    def add_item(cls, name, value):
        if not isinstance(value, (str, bool)):
            if isinstance(value, int):
                value = f'"{value}"'
            else:
                value = json.dumps(value)

        cls.run_command(Command.ADD_ITEM, name, value)

    @classmethod
    def delete(cls, name):
        try:
            cls.run_command(Command.DELETE, name)
        except Exception as e:  # pylint:disable=broad-except
            print("FAILED TO CLEAR STORAGE")
            print(e)

    @classmethod
    def clear(cls):
        try:
            cls.run_command(Command.CLEAR)
        except WebDriverException:
            print('Local storage is not accessible')

    @classmethod
    def add_encoded(cls, key, data):
        cls.add_item(
            key,
            base64.b64encode(json.dumps(data).encode('utf-8')).decode('utf-8')
        )

    @classmethod
    def verify_decoded(cls, key, expected_data):
        actual = cls.get_item(key)
        compare(
            expected=expected_data,
            actual=actual and json.loads(base64.b64decode(actual))
        )

    @classmethod
    def verify_item(cls, key, expected, rules=None):
        actual = cls.get_item(key)
        compare(
            expected=expected,
            actual=actual and json.loads(actual),
            rules=rules
        )



from pprint import pprint

from io import StringIO

from functools import wraps

import pytest

from fild_cfg import Cfg
from selenium.common.exceptions import WebDriverException

from fild_ui import folder
from fild_ui.browser import Browser
from fild_ui.screenshots import DownloadedImg, PageScreenshot


def is_screenshot_marked(item):
    for marker in item.own_markers:
        if marker.name == 'screenshot':
            return True

    return False


def save_screenshot_on_failure(outcome, item):

    report = outcome.get_result()

    if report.failed and Browser().is_open and not is_screenshot_marked(item):
        test_name = item.name

        if report.when == 'setup':
            test_name = f'setup_for_{test_name}'
        elif report.when != 'call':
            return

        file_name = folder.generate_file_name(test_name)
        file_path = folder.generate_file_path('screenshots', file_name)
        try:
            Browser().driver.save_screenshot(file_path)
        except WebDriverException:
            pass


def compare_screenshot(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        if Cfg.Screenshot.compare:
            PageScreenshot.initialize(f'{func.__module__}.{func.__name__}.png')

        func(*args, **kwargs)

        if Cfg.Screenshot.compare and not PageScreenshot.completed:
            PageScreenshot.compare(Browser().driver)

    return pytest.mark.screenshot(wrapped)


def compare_downloaded_img(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        if Cfg.Screenshot.compare:
            DownloadedImg.initialize(f'{func.__module__}.{func.__name__}.png')

        func(*args, **kwargs)

        if Cfg.Screenshot.compare:
            DownloadedImg.compare()

    return pytest.mark.screenshot(wrapped)


def get_console_errors():
    exclude = Cfg.Browser.get('exclude_logs_from') or []

    def is_excluded(log_item):
        for exclude_item in exclude:
            if exclude_item in log_item['message']:
                return True

        return False

    error_log = Browser().console_log

    return [log_item for log_item in error_log if not is_excluded(log_item)]


def check_console_errors(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        func(*args, **kwargs)
        error_log = get_console_errors()
        printed_log = StringIO()
        pprint(error_log, printed_log)
        assert not error_log, f'Console log errors.\n{printed_log.getvalue()}'

    return pytest.mark.extended(wrapped)



from selenium.common.exceptions import (
    ElementClickInterceptedException, StaleElementReferenceException,
)
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.support.select import Select as BaseSelect
from waiting import wait

from fild_ui.browser import Browser, Element, Elements


class Container(Element):
    pass


class Button(Element):
    @property
    def text(self):
        return self.located.text

    def click(self):
        self._target = None
        self.scroll_to()
        self.located.click()
        Browser().recreate_session()

        return True  # for sync in waits

    def click_once_loaded(self):
        return wait(
            self.click,
            timeout_seconds=2,
            sleep_seconds=0,
            expected_exceptions=(ElementClickInterceptedException,),
            waiting_for='clicking button once loaded'
        )

    def click_and_switch_to_new_window(self):
        windows = Browser().driver.window_handles
        self.click()
        wait(
            lambda: len(Browser().driver.window_handles) > len(windows),
            timeout_seconds=2,
            waiting_for='new window to open'
        )
        Browser().driver.switch_to.window(
            Browser().driver.window_handles[len(windows)]
        )

    @property
    def is_disabled(self):
        return bool(self.get_attribute('disabled'))

    def verify_text(self, expected):
        actual = self.text
        assert actual == expected, (
            f'Unexpected menu text.\nExpected: {expected}\nActual: {actual}'
        )

    def wait_for_text(self, expected_text, timeout_seconds=2):
        wait(
            lambda: expected_text == self.text,
            sleep_seconds=0,
            timeout_seconds=timeout_seconds,
            waiting_for=f'button text to equal {expected_text}'
        )


class Link(Button):
    @property
    def url(self):
        return self.located.get_attribute('href')


class TextInput(Element):
    def input(self, value):
        self.wait_for_load()

        self.located.click()
        self.located.clear()
        self.located.send_keys(value)

    def safe_input(self, value):
        self.wait_for_load()
        self.located.click()
        action = ActionChains(Browser().driver)
        action.send_keys(value)
        action.perform()

    def clear_and_type(self, value):
        self.wait_for_load()
        self.located.clear()
        self.safe_input(value)

    def get_value(self):
        self.wait_for_load()
        return self.located.get_attribute('value')

    def set_value(self, value):
        self.wait_for_load()
        Browser().driver.execute_script(
            'arguments[0].value = arguments[1]',
            self.located,
            value
        )

    def verify_value(self, expected):
        actual = self.get_value()
        assert actual == str(expected), (
            f'Unexpected input value.\nExpected: {expected}\nGot: {actual}'
        )


class Label(Element):
    @property
    def text(self):
        return self.located.text

    @property
    def is_displayed(self):
        return self.located.is_displayed

    def wait_for_updated(self, timeout=1):
        initial_text = self.located.text

        def _waiter():
            self.wait_for_load()
            return self.located.text != initial_text

        wait(_waiter, timeout_seconds=timeout)

    def wait_for_text(self, value, timeout=1):
        def _waiter():
            try:
                self.wait_for_load()
                return self.located.text == str(value)
            except StaleElementReferenceException:
                self.invalidate()
                return False

        wait(
            _waiter,
            timeout_seconds=timeout,
            waiting_for=f'waiting for text {value}'
        )

    def verify_label_text(self, text):
        self.wait_for_load()
        text = str(text)
        actual = self.located.text
        assert actual == text, (
            f'Unexpected label text.\nExpected: {text}\nGot: {actual}'
        )


class Select(Element):
    @property
    def located(self):
        return BaseSelect(super().located)

    def wait_for_value(self):
        def wait_for_options():
            self._target = None
            return self.located.options

        wait(
            wait_for_options,
            timeout_seconds=5,
            waiting_for='selected option to appear'
        )

    def select(self, value=None, index=None, text=None):
        if value is not None:
            self.located.select_by_value(value)
        elif index is not None:
            self.located.select_by_index(index)
        elif text is not None:
            self.located.select_by_visible_text(text)
        else:
            raise ValueError('provide value for select')

    @property
    def all_options(self):
        return [opt.text for opt in self.located.options]

    @property
    def selected_option(self):
        return self.located.first_selected_option.text

    def verify_selected_option(self, text):
        actual = self.selected_option
        assert actual == str(text), (
            f'Unexpected selected option.\nExpected: {text}\nGot: {actual}'
        )


class Checkbox(Button):
    @property
    def checked(self):
        return self.located.is_selected()

    def verify_checked(self, expected):
        actual = self.checked
        assert actual == expected, f'Expected checkbox to be: {expected}'


class TableRow(Container):
    Columns = Elements(css='td', element_class=Label)


class Table(Container):
    Head = Elements(css='th', element_class=Label)
    Rows = Elements(css='tbody tr', element_class=TableRow)

    def read_data(self):
        Browser().recreate_session()
        headers = [col.text for col in self.Head]
        data = [[col.text for col in row.Columns] for row in self.Rows]

        return headers, data



import os
import shutil

from fild_cfg import Cfg
from PIL import Image, ImageDraw
from waiting import wait

from fild_ui import folder


class CompareBase:
    def __init__(self, test_name):
        self.file_name = test_name

        self.failed_screenshot = folder.generate_file_path(
            'screenshots/failed', self.file_name
        )
        self.threshold_screenshot = folder.generate_file_path(
            'screenshots/threshold', self.file_name
        )

    def base_compare(self, expected, actual):
        expected_img = Image.open(expected)
        actual_img = Image.open(actual)

        screen_width, screen_height = actual_img.size
        failed_pixels = []

        for y in range(1, screen_height + 1):
            for x in range(1, screen_width + 1):
                region_actual = self.process_pixel(actual_img, x, y)
                region_expected = self.process_pixel(expected_img, x, y)

                if region_expected != region_actual:
                    failed_pixels.append((x, y))

        mismatch = (len(failed_pixels) / (screen_width * screen_height)) * 100

        if mismatch:
            print(f'Mismatch Percentage is: {mismatch}')
            draw = ImageDraw.Draw(actual_img)

            for pixel in failed_pixels:
                draw.rectangle([pixel, pixel], outline='red')

        if mismatch > Cfg.Screenshot.threshold:
            actual_img.save(self.failed_screenshot)
        if 0 < mismatch <= Cfg.Screenshot.threshold:
            actual_img.save(self.threshold_screenshot)

        return mismatch

    @staticmethod
    def process_pixel(image, x, y):
        """
        This can be used as the sensitivity factor,
        the larger it is the less sensitive the comparison
        """
        factor = 10

        try:
            pixel = image.getpixel((x, y))
            region_total = sum(pixel)/4
        except:  # pylint: disable=bare-except
            return None

        return region_total/factor


class Screenshot(CompareBase):
    def __init__(self, target, test_name):
        super().__init__(test_name=test_name)

        self.target = target

        self.expected_folder = test_name.split('.')[1].rstrip('_extra').rstrip(
            '_reports'
        )
        self.expected_file = '.'.join(test_name.split('.')[2:])

        self.actual_screenshot = folder.generate_file_path(
            'screenshots/actual', self.file_name
        )
        self.expected_screenshot = folder.generate_file_path(
            f'qa/screenshots/{self.expected_folder}', self.expected_file
        )

    def take(self):
        self.target.save_screenshot(self.actual_screenshot)
        self.target.get_screenshot_as_png()

    def generate_expected(self):
        self.target.save_screenshot(self.expected_screenshot)
        self.target.get_screenshot_as_png()

    def compare(self):
        try:
            Image.open(self.expected_screenshot)
        except FileNotFoundError as e:
            self.generate_expected()
            raise FileNotFoundError('Expected screenshot is missing.') from e

        return self.base_compare(
            expected=self.expected_screenshot,
            actual=self.actual_screenshot
        )


class PageScreenshot:
    name = None
    completed = False
    save_mode = False

    @staticmethod
    def set_save_mode(save_mode=False):
        PageScreenshot.save_mode = save_mode

    @staticmethod
    def initialize(name):
        PageScreenshot.name = name
        PageScreenshot.completed = False

    @staticmethod
    def complete():
        PageScreenshot.name = None
        PageScreenshot.completed = True

    @classmethod
    def compare(cls, target):
        screenshot = Screenshot(target, cls.name)

        if cls.save_mode:
            screenshot.generate_expected()
            mismatch = 0
        else:
            screenshot.take()
            mismatch = screenshot.compare()

        cls.complete()

        assert mismatch < Cfg.Screenshot.threshold, 'Screenshots don\'t match'


class DownloadedImg:
    name = None
    save_mode = False

    @staticmethod
    def set_save_mode(save_mode=False):
        DownloadedImg.save_mode = save_mode

    @staticmethod
    def initialize(name):
        downloads_folder = folder.generate_path('downloads').absolute()
        shutil.rmtree(downloads_folder, ignore_errors=True)
        DownloadedImg.name = name

    @staticmethod
    def get_downloaded_file_with_wait():
        downloads_folder = folder.generate_path('downloads').absolute()

        def wait_for_png():
            files = os.listdir(downloads_folder)

            for file_name in files:
                if file_name.endswith('.png'):
                    return True

            return False

        wait(
            wait_for_png,
            timeout_seconds=4,
            waiting_for='file to download'
        )

        for filename in os.listdir(downloads_folder):
            return str(os.path.join(downloads_folder, filename))

        return None  # TODO check if exception needed

    @staticmethod
    def complete():
        DownloadedImg.name = None

    @classmethod
    def compare(cls):
        compare_base = CompareBase(cls.name)

        expected_folder = cls.name.split('.')[1].rstrip(
            '_extra'
        ).rstrip('_reports')
        expected_file = '.'.join(cls.name.split('.')[2:])

        expected_img = folder.generate_file_path(
            f'qa/screenshots/{expected_folder}/files', expected_file
        )
        actual_img = folder.generate_file_path(
            'screenshots/files', expected_file
        )

        if cls.save_mode:
            shutil.copy(
                cls.get_downloaded_file_with_wait(),
                expected_img
            )
            mismatch = 0
        else:
            mismatch = compare_base.base_compare(
                expected=expected_img,
                actual=cls.get_downloaded_file_with_wait()
            )
            # Saving file for publishing actual resul in case of failure
            shutil.copy(
                cls.get_downloaded_file_with_wait(),
                actual_img
            )

        cls.complete()

        assert mismatch < Cfg.Screenshot.threshold, 'Files don\'t match'
